/*
Copyright (c) 2004 Joseph Gleason

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Current versions of this and other code can be downloaded at:
http://gleason.cc/

*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "hash_table.h"
#include "hash_function.h"


//Most comments are in header file

//Internal functions
int hash_table_getslot(hash_table_t *T,void* Key, size_t KeyLen)
{
   size_t HashVal=hash(Key,KeyLen,0);
   size_t Slot=HashVal & (T->TableLength-1);
   return Slot;
}


//For debugging only
//counts number of entries in a slot
int hash_table_countslot(hash_table_slot_t *S)
{
   int Cnt=0;

   hash_table_entry_t *Curr=S->First;
   while(Curr!=NULL)
   {
      Cnt++;
      Curr=Curr->SlotNext;
   }

   return Cnt;
}



//External functions
void hash_table_init(hash_table_t *T, int InitalSize, double LoadFactor)
{
   size_t Sz=1;
   while (Sz<InitalSize) Sz*=2;

   T->TableLength=Sz;
   T->TableTotalElements=0;
   T->TableSlotsInUse=0;

   T->InsertFirst=NULL;
   T->InsertLast=NULL;
   T->Array=(hash_table_slot_t*)malloc(sizeof(hash_table_slot_t)*Sz);

   size_t i=0;
   for(i=0; i<Sz; i++)
   {
      T->Array[i].First=NULL;
   }

}



void hash_table_destroy(hash_table_t *T)
{


}

//Accessors
hash_table_entry_t* hash_table_getfirst(hash_table_t *T)
{
   return T->InsertFirst;
}
hash_table_entry_t* hash_table_getlast(hash_table_t *T)
{
   return T->InsertLast;
}
hash_table_entry_t* hash_table_get(hash_table_t *T,void* Key, size_t KeyLen)
{
   size_t Slot=hash_table_getslot(T,Key,KeyLen);

   hash_table_entry_t* Curr;

   Curr=T->Array[Slot].First;

   while(9)
   {
      if (Curr==NULL) return NULL;
      if ((Curr->KeyLen==KeyLen) && (memcmp(Key,Curr->Key,KeyLen)==0))
      {
         return Curr;
      }
      Curr=Curr->SlotNext;
   }
}

//Modifiers
int hash_table_put(hash_table_t *T,void* Key, size_t KeyLen, void* Data)
{
   //printf("Hashtable put: %lx\n",(uint64_t)Data);

   size_t Slot=hash_table_getslot(T,Key,KeyLen);


   hash_table_entry_t** Curr=&(T->Array[Slot].First);

   //If slot is empty, we will be filling it
   if (*Curr==NULL) T->TableSlotsInUse++;

   while(10)
   {
      if (*Curr==NULL)
      {
         *Curr=(hash_table_entry_t*)malloc(sizeof(hash_table_entry_t));
         hash_table_entry_t *Target=*Curr;
         Target->Key=malloc(KeyLen);
         memcpy(Target->Key,Key,KeyLen);
         Target->KeyLen=KeyLen;
         Target->Data=Data;
         Target->SlotNext=NULL;
         Target->InsertNext=NULL;
         Target->InsertPrev=T->InsertLast;
         assert(Target!=Target->InsertNext);

         if (T->InsertLast!=NULL)
         {
            T->InsertLast->InsertNext=Target;
         }

         if (T->InsertFirst==NULL) T->InsertFirst=Target;
         T->InsertLast=Target;
         T->TableTotalElements++;

         assert(*Curr!=NULL);

         assert(Target!=Target->InsertNext);
         assert(Target!=Target->InsertPrev);
         //printf("Hashtable put: %lx %lx\n",(uint64_t)Data,(uint64_t)(*Curr)->Data);
         return 1;
      }
      if (((*Curr)->KeyLen==KeyLen) && (memcmp(Key,(*Curr)->Key,KeyLen)==0))
      {
         (*Curr)->Data=Data;
         return 2;
      }

      Curr=&((*Curr)->SlotNext);

   }
}

void* hash_table_erase(hash_table_t *T,void* Key, size_t KeyLen)
{
   size_t SlotIdx=hash_table_getslot(T,Key,KeyLen);

   hash_table_slot_t* Slot=&(T->Array[SlotIdx]);

   //printf("Slot: %u %lx\n",SlotIdx,Slot);


   //printf("TopC\n");
   int CntInit=hash_table_countslot(Slot);
   hash_table_entry_t** Curr=&(Slot->First);



   while(31)
   {
      if (*Curr==NULL) return NULL; //Target not present

      //if found target
      if (((*Curr)->KeyLen==KeyLen) && (memcmp(Key,(*Curr)->Key,KeyLen)==0))
      {
         hash_table_entry_t* Target=*Curr;
         *Curr=Target->SlotNext; //Slot Parent now points to next entry

         if (CntInit==1) assert(*Curr==NULL);
         if (CntInit==1) assert(Slot->First==NULL);


         //If there is a valid guy in front of me
         //then I'll make it point to the clown behind me
         if (Target->InsertPrev!=NULL)
         {
            assert(Target->InsertPrev!=Target->InsertPrev->InsertNext);
            Target->InsertPrev->InsertNext=Target->InsertNext;
            assert(Target->InsertPrev!=Target->InsertPrev->InsertNext);
         }

         //If there is a valid guy behind me,
         //make it point to clown in front of me
         if (Target->InsertNext!=NULL)
         {
            assert(Target->InsertNext!=Target->InsertNext->InsertPrev);
            Target->InsertNext->InsertPrev=Target->InsertPrev;
            assert(Target->InsertNext!=Target->InsertNext->InsertPrev);

         }

         //If I am first or last in entire list,
         //update those pointers appropriately.
         if (T->InsertFirst==Target) T->InsertFirst=Target->InsertNext;
         if (T->InsertLast==Target) T->InsertLast=Target->InsertPrev;

         void* Data;
         Data=Target->Data;

         free(Target->Key);
         free(Target);

         T->TableTotalElements--;

         if(Slot->First==NULL)
            T->TableSlotsInUse--;



         //printf("BotC\n");
         assert(CntInit-1==hash_table_countslot(Slot));

         return Data;
      }

      //else case, go to next slot entry
      Curr=&((*Curr)->SlotNext);

   }


}
